#include <SPI.h>
#include <SD.h>
#include <Streaming.h>
#include "db.h"

// SS-pin and files
uint8_t ssPin = D0;              // SS pin SD 
char dbName[] =  "demo.db";      // file name database
char cdbName[] = "demo.cdb";     // file name database backup
// record data: only chars, convert nunbers to char*
const uint8_t dataW = 50;        // max data width in a record, exclusive 'key' and 'tombstone' marker

// local used //
char dataR[dataW + 1];  // record holder (dataW + 1 chars)
char key[kyW + 1];      // key holder (exactly kyW(=3) + 1 chars - kyW in 'db.h')
// DEMO TEMP data field names, JSON style: name|value (you can change the field names for each record)
char tempFields[][8] = {"tempIn", "tempOut", "time"};
// DEMO TEMP data field used as DB table style and as JSON style
float    tempIn    = 20.8;     // indoor temperature
float    tempOut   =  9.3; // outdoor temperature
uint32_t timestamp = 1423567890; // time temperature

void setup() {
  Serial.begin(115200);
  Serial << "\n\n";

  pinMode(ssPin, OUTPUT);
  SPI.setClockDivider(SPI_CLOCK_DIV2);
  if (!SD.begin(ssPin)) {Serial << "SD not ready, STOP\n"; return;}

  test();
}

void test() {
  showFiles();  // db files on SD?
  /********* fresh start? remove // on next 2 line to delete the db *********/
  Serial << "delete database...\n";
  if (SD.exists(dbName)) if (!SD.remove(dbName)) Serial << "db file not deleted";  // delete file error
  /**************************************************************************/
 
  // filenames: short natation max 8.3 'naaaaame.ext'
  initDb(ssPin, dbName, cdbName, dataW);  // SSPin, db name, copy/backup name, max data width
  if (dbErr) {Serial << "initDB ERROR, STOP\n"; return;}
  if (!SD.exists(dbN)) {Serial << "exists ERROR, STOP\n"; return;}
  nrRec();  // get the number of records in db
  if (dbErr) {Serial << "nrRec ERROR, STOP\n"; return;}

  // now, we can use the database
  Serial << "number of records: " << dbR << (dbR == 0 ? "\tDatabase is empty!" : "")<< endl;
  
  // ##### if empty: add a few records #####
  if (dbR == 0) {  // no records, fill the DB with a few records
    // add a record in JSON style
    strcpy(key, "T00");  // Upercase T(emp)
    // as JSON style
    // format data: 3 x fieldname|value| (seperated and ended with a '|')
    sprintf(dataR, "%s|%s|%s|%s|%s|%d|",
        tempFields[0], dtostr(tempIn, 5, 2, ""), 
        tempFields[1], dtostr(tempOut, 5, 2, ""), 
        tempFields[2], timestamp);
    addRec(key, dataR);
    Serial << "add: T00";
    showErr();
    Serial << endl;
  }
  if (dbR == 1) {  // add a record in DB table style (the same data as in the previous record) 
    // format data: 3 x value| (seperated and ended with a '|')
    strcpy(key, "t01");  // lowercase t(emp)
    sprintf(dataR, "%s|%s|%d|",
        dtostr(tempIn, 5, 2, ""),
        dtostr(tempOut, 5, 2, ""),
        timestamp);
    addRec(key, dataR);
    Serial << "add: t01";
    showErr();
    Serial << endl;
  }
  if (dbR == 2) {  // add a record in JSON style
    enum {OFF, ON};  // DEMO enum for Room lamps: OFF or ON
    // format data: 4 x fieldname|value| (seperated and ended with a '|')
    sprintf(dataR, "%s|%s|%s|%d|%s|%d|%s|%d|",
        "room", "Living",
        "lamp1", OFF,  // is "0"
        "lamp2", ON,   // is "1"
        "dimLevel2", 5);
    strcpy(key, "R00");  // key: R(oom) #00
    addRec(key, dataR);
    Serial << "add: R00";
    showErr();
    Serial << endl;
  }
  Serial << "\nnumber of records: " << dbR << endl;

  // ##### overwrire R00 #####
  enum {OFF, ON};  // DEMO enum for Room lamps: OFF or ON
  // format data: 4 x fieldname|value| (seperated and ended with a '|')
  sprintf(dataR, "%s|%s|%s|%d|%s|%d|%s|%d|",
      "room", "Living",
      "lamp1", ON,  
      "lamp2", OFF,  
      "dimLevel2", 5);
  strcpy(key, "R00");
  writeRec(key, dataR);
  Serial << "\noverwrite: R00";
  showErr();
  Serial << endl;
  
  // ##### list keys #####
  Serial << "\nlist keys:\n";
  listRec(false);
  showErr();

  // ##### list keys with data #####
  Serial << "\nlist keys with data:\n";
  listRec(true);
  showErr();

  // ##### This record does not exist #####
  Serial << "\nfind 't00' (does not exist):\n";
  if (findRec((char*)"t00")) Serial << "'t00' exist, position: " << dbF.position() << endl;
  else Serial << "\t't00' not found!\n";

  // ##### This record must exist #####
  Serial << "\nfind 't01' (must exist):\n";
  if (findRec((char*)"t01")) Serial << "\t't01' exist, position: " << dbF.position() << endl;
  else Serial << "\t't01' not found!\n";

  // ##### These three records must exist #####
  Serial << "\nread rec 't01' (rec #2):\n";
  readRec((char*)"t01");
  if (dbErr) {Serial << "\t't01' add error! "; showErr();}
  else {
    Serial << dbB << endl;
    parseRec(dbB, false);
  }

  Serial << "\nread rec 'T00' (rec #1):\n";
  readRec((char*)"T00");
  if (dbErr) {Serial << "\t'T00' error: "; showErr();}
  else {
    Serial << dbB << endl;
    parseRec(dbB, true);
  }

  Serial << "\nread rec 'R00' (rec #3):\n";
  readRec((char*)"R00");
  if (dbErr) {Serial << "\t'R00' error: "; showErr();}
  else {
    Serial << dbB << endl;
    parseRec(dbB, true);
  }
  
  // ##### tombstone records #####
  Serial << "\ntombstone 't00' (does not exist):\n";
  delRec((char*)"t00"); 
  if (dbErr) {Serial << "\t't00' del error: "; showErr();}

  Serial << "\ntombstone 't01' (rec #2):\n";
  delRec((char*)"t01"); 
  if (dbErr) {Serial << "\t't01' del error: "; showErr();}

  // ##### This record is a tombstone #####
  Serial << "\nfind 't01' (tombstone):\n";
  if (findRec((char*)"t01")) Serial << "\t't01' exist, position: " << dbF.position() << endl;
  else Serial << "\t't01' not found!\n";

  dbF.close();
  
  // ##### database: compress and backup #####
  Serial << "\nDB copy and compress\n\n";
  cmprDb();
  if (dbErr) {Serial << "resul: "; showErr();}
  dbF.close();  

  showFiles();  // db files on SD
}

void parseRec(char *s, boolean f) {  // s: record, f: false=value style, true=names|value style
  char *n, *v, *b;
  int w = strlen(s) - 1;
  while (s[--w] == ' ') s[w] = '\0';   // remove trailing white-space (added by addRec())

  n = strtok_r(s, "|", &b);
  if (f) {
    do {
      v = strtok_r(NULL, "|", &b);
      Serial << "field:value  " << n << ":" << v << endl;
    } while (n = strtok_r(NULL, "|", &b));
  }
  else {
    while (n = strtok_r(NULL, "|", &b)) Serial << "value: " << n << endl;
  }
}

void showFiles() {  // only the file(s) of this db
  Serial << "==== db files ====\n";
  // file names to uppercae
  char dName[strlen(dbName) + 1];
  char cName[strlen(cdbName) + 1];
  dName[strlen(dbName)] = '\0';
  cName[strlen(cdbName)] = '\0';
  for (uint8_t i = 0; i < strlen(dbName); i++) dName[i] = toupper(dbName[i]);
  for (uint8_t i = 0; i < strlen(cdbName); i++) cName[i] = toupper(cdbName[i]);

  char f[13];
  File dir = SD.open("/");
  if (!dir) {Serial << "showFiles(): SD.open() error\n"; return;}
  dir.rewindDirectory();
  while (true) {
    File entry =  dir.openNextFile();
    if (! entry) break;  // no more files
    if (!entry.isDirectory()) {
      if (strcmp(entry.name(), dName) == 0 || strcmp(entry.name(), cName) == 0) {
        sprintf(f, "%-12s  ", entry.name() );
        Serial << f << entry.size() << endl;
      }
    }
    entry.close();
  }
  dir.close();
  Serial << "==================\n";
}

void loop() {
  ;
}


